from collections import deque
from io import BytesIO

try:
    from typing import Optional, Union
except ImportError:
    pass

try:
    import stringzilla as sz
except ImportError:
    sz = None

from . import ExcessBitsError, bit_size, compute_min_pattern_size, initialize_dictionary

# encodes [min_pattern_bytes, min_pattern_bytes + 13] pattern lengths
_huffman_codes = b"\x00\x03\x08\x0b\x14$&+KT\x94\x95\xaa'"
# These bit lengths pre-add the 1 bit for the 0-value is_literal flag.
_huffman_bits = b"\x02\x03\x05\x05\x06\x07\x07\x07\x08\x08\x09\x09\x09\x07"
_FLUSH_CODE = 0xAB  # 8 bits


class _BitWriter:
    """Writes bits to a stream."""

    def __init__(self, f, close_f_on_close=False):
        self.close_f_on_close = close_f_on_close
        self.f = f
        self.buffer = 0  # Basically a uint24
        self.bit_pos = 0

    def write_huffman(self, pattern_size):
        return self.write(_huffman_codes[pattern_size], _huffman_bits[pattern_size])

    def write(self, bits, num_bits, flush=True):
        bits &= (1 << num_bits) - 1
        self.bit_pos += num_bits
        self.buffer |= bits << (32 - self.bit_pos)

        bytes_written = 0
        if flush:
            while self.bit_pos >= 8:
                byte = self.buffer >> 24
                self.f.write(byte.to_bytes(1, "big"))
                self.buffer = (self.buffer & 0xFFFFFF) << 8
                self.bit_pos -= 8
                bytes_written += 1
        return bytes_written

    def flush(self, write_token=True):
        bytes_written = 0
        if self.bit_pos > 0 and write_token:
            bytes_written += self.write(_FLUSH_CODE, 9)

        while self.bit_pos > 0:
            byte = (self.buffer >> 24) & 0xFF
            self.f.write(byte.to_bytes(1, "big"))
            self.bit_pos = 0
            self.buffer = 0
            bytes_written += 1

        self.f.flush()

        return bytes_written

    def close(self):
        self.flush(write_token=False)
        if self.close_f_on_close:
            self.f.close()


class _RingBuffer:
    def __init__(self, buffer):
        self.buffer = buffer
        self.size = len(buffer)
        self.pos = 0  # Always pointing to the byte-to-be-overwritten

    def write_byte(self, byte):  # ~10% of time
        self.buffer[self.pos] = byte
        self.pos = (self.pos + 1) % self.size

    def write_bytes(self, data):
        for byte in data:
            self.write_byte(byte)

    def index(self, pattern, start):
        result = sz.find(self.buffer, pattern, start=start) if sz else self.buffer.find(pattern, start)

        if result == -1:
            raise ValueError("substring not found")
        return result


class Compressor:
    """Compresses data to a file or stream."""

    def __init__(
        self,
        f,
        *,
        window: int = 10,
        literal: int = 8,
        dictionary: Optional[bytearray] = None,
        lazy_matching: bool = False,
    ):
        """
        Parameters
        ----------
        f: Union[str, Path, FileLike]
            Path/FileHandle/Stream to write compressed data to.
        window: int
            Size of window buffer in bits.
            Higher values will typically result in higher compression ratios and higher computation cost.
            A same size buffer is required at decompression time.
            Valid range: ``[8, 15]``.
            Defaults to ``10`` (``1024`` byte buffer).
        literal: int
            Number of used bits in each byte of data.
            The default ``8`` bits can store all data.
            A common other value is ``7`` for storing ascii characters where the most-significant-bit is always 0.
            Smaller values result in higher compression ratios for no additional computation cost.
            Valid range: ``[5, 8]``.
        dictionary: Optional[bytearray]
            Use the given **initialized** buffer inplace.
            At decompression time, the same initialized buffer must be provided.
            ``window`` must agree with the dictionary size.
            If providing a pre-allocated buffer, but with default initialization, it must
            first be initialized with :func:`~tamp.initialize_dictionary`
        lazy_matching: bool
            Use roughly 50% more cpu to get 0~2% better compression.
        """
        if lazy_matching:
            raise NotImplementedError("lazy matching not implemented in pure python implementation.")

        if not hasattr(f, "write"):  # It's probably a path-like object.
            # TODO: then close it on close
            f = open(str(f), "wb")
            close_f_on_close = True
        else:
            close_f_on_close = False

        self._bit_writer = _BitWriter(f, close_f_on_close=close_f_on_close)
        if dictionary and bit_size(len(dictionary) - 1) != window:
            raise ValueError("Dictionary-window size mismatch.")

        self.window_bits = window
        self.literal_bits = literal

        self.min_pattern_size = compute_min_pattern_size(window, literal)
        self.max_pattern_size = self.min_pattern_size + 13

        self.literal_flag = 1 << self.literal_bits

        self._window_buffer = _RingBuffer(
            buffer=dictionary if dictionary else initialize_dictionary(1 << window),
        )

        self._input_buffer = deque(maxlen=self.max_pattern_size)

        # Callbacks for debugging/metric collection; can be externally set.
        self.token_cb = None
        self.literal_cb = None
        self.flush_cb = None

        # Write header
        self._bit_writer.write(window - 8, 3, flush=False)
        self._bit_writer.write(literal - 5, 2, flush=False)
        self._bit_writer.write(bool(dictionary), 1, flush=False)
        self._bit_writer.write(0, 1, flush=False)  # Reserved
        self._bit_writer.write(0, 1, flush=False)  # No other header bytes

    def _compress_input_buffer_single(self) -> int:
        target = bytes(self._input_buffer)
        bytes_written = 0
        search_i = 0
        match_size = 1
        for match_size in range(self.min_pattern_size, len(target) + 1):
            match = target[:match_size]
            try:
                search_i = self._window_buffer.index(match, search_i)
            except ValueError:
                # Not Found
                match_size -= 1
                break
        match = target[:match_size]

        if match_size >= self.min_pattern_size:
            if self.token_cb:
                self.token_cb(
                    search_i,
                    match_size,
                    match,
                )
            bytes_written += self._bit_writer.write_huffman(match_size - self.min_pattern_size)
            bytes_written += self._bit_writer.write(search_i, self.window_bits)
            self._window_buffer.write_bytes(match)

            for _ in range(match_size):
                self._input_buffer.popleft()
        else:
            char = self._input_buffer.popleft()
            if self.literal_cb:
                self.literal_cb(char)
            if char >> self.literal_bits:
                raise ExcessBitsError

            bytes_written += self._bit_writer.write(char | self.literal_flag, self.literal_bits + 1)
            self._window_buffer.write_byte(char)

        return bytes_written

    def write(self, data: Union[bytes, bytearray]) -> int:
        """Compress ``data`` to stream.

        Parameters
        ----------
        data: Union[bytes, bytearray]
            Data to be compressed.

        Returns
        -------
        int
            Number of compressed bytes written.
            May be zero when data is filling up internal buffers.
        """
        bytes_written = 0

        for char in data:
            self._input_buffer.append(char)
            if len(self._input_buffer) == self._input_buffer.maxlen:
                bytes_written += self._compress_input_buffer_single()

        return bytes_written

    def flush(self, write_token: bool = True) -> int:
        """Flushes all internal buffers.

        This compresses any data remaining in the input buffer,
        and flushes any remaining data in the output buffer to
        disk.

        Parameters
        ----------
        write_token: bool
            If appropriate, write a ``FLUSH`` token.
            Defaults to :obj:`True`.

        Returns
        -------
        int
            Number of compressed bytes flushed to disk.
        """
        bytes_written = 0
        if self.flush_cb:
            self.flush_cb()
        while self._input_buffer:
            bytes_written += self._compress_input_buffer_single()
        bytes_written += self._bit_writer.flush(write_token=write_token)
        return bytes_written

    def close(self) -> int:
        """Flushes all internal buffers and closes the output file or stream, if tamp opened it.

        Returns
        -------
        int
            Number of compressed bytes flushed to disk.
        """
        bytes_written = 0
        bytes_written += self.flush(write_token=False)
        self._bit_writer.close()
        return bytes_written

    def __enter__(self) -> "Compressor":
        """Use :class:`Compressor` as a context manager.

        .. code-block:: python

           with tamp.Compressor("output.tamp") as f:
               f.write(b"foo")
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Calls :meth:`~Compressor.close` on contextmanager exit."""
        self.close()


class TextCompressor(Compressor):
    """Compresses text to a file or stream."""

    def write(self, data: str) -> int:
        return super().write(data.encode())


def compress(
    data: Union[bytes, str],
    *,
    window: int = 10,
    literal: int = 8,
    dictionary: Optional[bytearray] = None,
    lazy_matching: bool = False,
) -> bytes:
    """Single-call to compress data.

    Parameters
    ----------
    data: Union[str, bytes]
        Data to compress.
    window: int
        Size of window buffer in bits.
        Higher values will typically result in higher compression ratios and higher computation cost.
        A same size buffer is required at decompression time.
        Valid range: ``[8, 15]``.
        Defaults to ``10`` (``1024`` byte buffer).
    literal: int
        Number of used bits in each byte of data.
        The default ``8`` bits can store all data.
        A common other value is ``7`` for storing ascii characters where the most-significant-bit is always 0.
        Valid range: ``[5, 8]``.
    dictionary: Optional[bytearray]
        Use the given **initialized** buffer inplace.
        At decompression time, the same initialized buffer must be provided.
        ``window`` must agree with the dictionary size.
        If providing a pre-allocated buffer, but with default initialization, it must
        first be initialized with :func:`~tamp.initialize_dictionary`
    lazy_matching: bool
        Use roughly 50% more cpu to get 0~2% better compression.

    Returns
    -------
    bytes
        Compressed data
    """
    with BytesIO() as f:
        if isinstance(data, str):
            c = TextCompressor(
                f,
                window=window,
                literal=literal,
                dictionary=dictionary,
                lazy_matching=lazy_matching,
            )
            c.write(data)
        else:
            c = Compressor(
                f,
                window=window,
                literal=literal,
                dictionary=dictionary,
                lazy_matching=lazy_matching,
            )
            c.write(data)
        c.flush(write_token=False)
        f.seek(0)
        return f.read()



import os
import shutil
from pathlib import Path

allowed_to_fail = os.environ.get("CIBUILDWHEEL", "0") != "1"

profile = os.environ.get("TAMP_PROFILE", "0") == "1"

# Enable sanitizers by default for development builds, disable for CI/production
in_ci = os.environ.get("CI", "false").lower() == "true"
building_wheel = os.environ.get("CIBUILDWHEEL", "0") == "1"
is_production_build = in_ci or building_wheel

sanitize = os.environ.get("TAMP_SANITIZE", "0") == "1"

# Ensure profile and sanitize are mutually exclusive
if profile and sanitize:
    raise ValueError(
        "Cannot enable both profiling and sanitizers at the same time. "
        "Please choose either TAMP_PROFILE=1 or TAMP_SANITIZE=1, not both."
    )


def build_cython_extensions():
    import Cython.Compiler.Options
    from Cython.Build import build_ext, cythonize
    from Cython.Compiler.Options import get_directive_defaults
    from setuptools import Extension
    from setuptools.dist import Distribution

    Cython.Compiler.Options.annotate = True

    if sanitize:
        print("Building with sanitizers enabled (UBSan + ASan)")
    elif is_production_build:
        print("Building for production (sanitizers disabled)")
    else:
        print("Building for development")

    define_macros = []

    define_macros.append(("TAMP_LAZY_MATCHING", "1"))

    # Force embedded find_best_match implementation on desktop (for testing)
    if os.environ.get("TAMP_USE_EMBEDDED_MATCH", "0") == "1":
        print("Using embedded find_best_match implementation")
        define_macros.append(("TAMP_USE_EMBEDDED_MATCH", "1"))

    if profile:
        print("Setting profiling configuration.")
        directive_defaults = get_directive_defaults()
        directive_defaults["linetrace"] = True
        directive_defaults["binding"] = True

        define_macros.append(("CYTHON_TRACE", "1"))

    if os.name == "nt":  # Windows
        if profile:  # noqa: SIM108
            extra_compile_args = [
                "/O2",
                "/Z7",  # Debug info for profiling
            ]
        else:
            extra_compile_args = [
                "/O2",
            ]
    else:  # UNIX-based systems
        if sanitize:
            extra_compile_args = [
                "-O0",  # No optimization for better sanitizer output
                "-g",  # Include debug symbols
                "-fsanitize=undefined",
                "-fsanitize=address",
                "-fno-omit-frame-pointer",
                "-Wno-unreachable-code-fallthrough",
                "-Wno-deprecated-declarations",
                "-Wno-parentheses-equality",
                "-Wno-unreachable-code",
            ]
            extra_link_args = ["-fsanitize=undefined", "-fsanitize=address"]
        elif profile:
            extra_compile_args = [
                "-O2",  # Use O2 instead of O3 for better debug info
                "-g",  # Include debug symbols
                "-fno-omit-frame-pointer",
                "-Wno-unreachable-code-fallthrough",
                "-Wno-deprecated-declarations",
                "-Wno-parentheses-equality",
                "-Wno-unreachable-code",
            ]
            extra_link_args = []
        else:
            extra_compile_args = [
                "-O3",
                "-Werror",
                "-fno-omit-frame-pointer",  # counterintuitively, makes things faster?
                "-Wno-unreachable-code-fallthrough",
                "-Wno-deprecated-declarations",
                "-Wno-parentheses-equality",
                "-Wno-unreachable-code",  # TODO: This should no longer be necessary with Cython>=3.0.3
                # https://github.com/cython/cython/issues/5681
            ]
            extra_link_args = []
    include_dirs = ["tamp/_c_src/", "tamp/"]

    extension_kwargs = {
        "include_dirs": include_dirs,
        "extra_compile_args": extra_compile_args,
        "language": "c",
        "define_macros": define_macros,
    }

    if sanitize:
        extension_kwargs["extra_link_args"] = extra_link_args

    extensions = [
        Extension(
            "tamp._c_compressor",
            [
                "tamp/_c_src/tamp/common.c",
                "tamp/_c_src/tamp/compressor.c",
                "tamp/_c_compressor.pyx",
            ],
            **extension_kwargs,
        ),
        Extension(
            "tamp._c_decompressor",
            [
                "tamp/_c_src/tamp/common.c",
                "tamp/_c_src/tamp/decompressor.c",
                "tamp/_c_decompressor.pyx",
            ],
            **extension_kwargs,
        ),
        Extension(
            "tamp._c_common",
            [
                "tamp/_c_common.pyx",
            ],
            **extension_kwargs,
        ),
    ]

    include_dirs = set()
    for extension in extensions:
        include_dirs.update(extension.include_dirs)
    include_dirs = list(include_dirs)

    # Configure cythonize options for profiling
    cythonize_kwargs = {
        "include_path": include_dirs,
        "language_level": 3,
        "annotate": True,
    }

    if profile:
        # Enable additional debugging for profiling
        cythonize_kwargs["gdb_debug"] = True
        cythonize_kwargs["emit_linenums"] = True

    ext_modules = cythonize(extensions, **cythonize_kwargs)
    dist = Distribution({"ext_modules": ext_modules})
    cmd = build_ext(dist)
    cmd.ensure_finalized()
    cmd.run()

    for output in cmd.get_outputs():
        output = Path(output)
        relative_extension = output.relative_to(cmd.build_lib)
        shutil.copyfile(output, relative_extension)


if os.environ.get("TAMP_BUILD_C_EXTENSIONS", "1") == "1":
    try:
        build_cython_extensions()
    except Exception:
        if not allowed_to_fail:
            raise



from io import BytesIO

import micropython
from micropython import const

from . import compute_min_pattern_size, initialize_dictionary

_HUFFMAN_TABLE = b"BBBBBBBBBBBBBBBBeeee\x8a\x8bxxffffmmmmTTTTTTTTyy\x8c\x8fggggCCCCCCCCCCCCCCCC"

_FLUSH = const(15)


class Decompressor:
    def __init__(self, f, *, dictionary=None):
        self._close_f_on_close = False
        if not hasattr(f, "read"):  # It's probably a path-like object.
            f = open(str(f), "rb")
            self._close_f_on_close = True

        self.f = f
        self.f_buf = 0
        self.f_pos = 0

        # Read Header
        header = self.f.read(1)[0]
        self.w_bits = (header >> 5) + 8
        self.literal_bits = ((header >> 3) & 0b11) + 5
        uses_custom_dictionary = header & 0b100
        reserved = header & 0b10
        more_header_bytes = header & 0b1

        if reserved or more_header_bytes:
            raise NotImplementedError

        if uses_custom_dictionary:
            if not dictionary:
                raise ValueError
            self.w_buf = dictionary
        else:
            self.w_buf = initialize_dictionary(1 << self.w_bits)
        self.w_pos = 0

        self.min_pattern_size = compute_min_pattern_size(self.w_bits, self.literal_bits)

        self.overflow_buf = bytearray(self.min_pattern_size + 13)
        self.overflow_pos = 0
        self.overflow_size = 0

    @micropython.viper
    def readinto(self, buf) -> int:
        """Attempt to fill out, will place into overflow."""
        out_capacity = int(len(buf))  # const
        out_buf = ptr8(buf)

        huffman_table = ptr8(_HUFFMAN_TABLE)

        literal_bits = int(self.literal_bits)

        overflow_buf = self.overflow_buf
        overflow_buf_ptr = ptr8(overflow_buf)
        overflow_pos = int(self.overflow_pos)
        overflow_size = int(self.overflow_size)

        w_buf = self.w_buf
        w_buf_ptr = ptr8(w_buf)
        w_pos = int(self.w_pos)
        w_bits = int(self.w_bits)
        w_mask = (1 << w_bits) - 1

        f_pos = int(self.f_pos)
        f_buf = int(self.f_buf)
        f = self.f

        min_pattern_size = int(self.min_pattern_size)
        int(len(overflow_buf))

        # Copy overflow into out if available
        overflow_copy = int(min(overflow_size, out_capacity))
        for i in range(overflow_copy):
            buf[i] = overflow_buf_ptr[overflow_pos + i]
        out_pos = overflow_copy
        overflow_pos += overflow_copy
        overflow_size -= overflow_copy

        full_mask = int(0x3FFFFFFF)

        # Decompress more
        try:
            while out_pos < out_capacity:
                overflow_pos = 0  # overflow_size is also always zero here

                if f_pos == 0:
                    f_buf = int(f.read(1)[0]) << 22  # Will raise IndexError if out of data.
                    f_pos = 8

                # reuse is_literal flag as match_size so we don't need to explicitly set it
                match_size = f_buf >> 29
                # Don't update f_buf & f_pos here, in case there's not enough data available.

                if match_size:
                    if f_pos < (literal_bits + 1):
                        f_buf |= int(f.read(1)[0]) << (22 - f_pos)
                        f_pos += 8

                    # Now update buffers from is_literal
                    f_buf = (f_buf << 1) & full_mask

                    c = f_buf >> (30 - literal_bits)
                    f_buf = (f_buf << literal_bits) & full_mask
                    f_pos -= literal_bits + 1

                    out_buf[out_pos] = c
                    w_buf_ptr[w_pos] = c
                    out_pos += 1
                else:
                    # Read and decode Huffman; we'll need the bits regardless
                    if f_pos < 9:
                        f_buf |= int(f.read(1)[0]) << (22 - f_pos)
                        f_pos += 8

                    if f_buf >> 28:
                        code = (f_buf >> 21) & 0x7F
                        code = 33 if code >= 64 else huffman_table[code]
                        match_size = code & 0xF
                        if match_size == _FLUSH:
                            f_buf = f_pos = 0
                            continue
                        delta = code >> 4
                    else:
                        match_size = 0
                        delta = 1

                    token_bits = 1 + delta + w_bits

                    while f_pos < token_bits:
                        f_buf |= int(f.read(1)[0]) << (22 - f_pos)
                        f_pos += 8

                    # We now absolutely have enough data to decode token.

                    match_size += min_pattern_size
                    index = (f_buf >> (30 - token_bits)) & w_mask
                    f_buf = (f_buf << token_bits) & full_mask
                    f_pos -= token_bits

                    # Copy entire match to overflow
                    for i in range(match_size):
                        overflow_buf_ptr[i] = w_buf_ptr[index + i]
                    overflow_pos = 0
                    overflow_size = match_size

                    # Copy available amount to out_buf
                    for i in range(match_size):
                        if out_pos < out_capacity:
                            out_buf[out_pos] = overflow_buf_ptr[overflow_pos]
                            out_pos += 1
                            overflow_pos += 1
                            overflow_size -= 1
                        # Copy overflow to window
                        w_buf_ptr[(w_pos + i) & w_mask] = overflow_buf_ptr[i]
                w_pos = (w_pos + match_size) & w_mask
        except IndexError:
            pass

        self.w_pos = w_pos
        self.overflow_pos = overflow_pos
        self.overflow_size = overflow_size
        self.f_buf = f_buf
        self.f_pos = f_pos

        return out_pos

    def read(self, size=-1):
        """Return at most ``size`` bytes."""
        if size == 0:
            out = bytearray()
        elif size < 0:
            # Read into a bunch of chunks, then do a single join.
            chunks = []
            chunk_size = 1024
            decompressed_bytes = 0
            while True:
                chunk = bytearray(chunk_size)
                chunk_decompressed_bytes = self.readinto(chunk)
                if chunk_decompressed_bytes != chunk_size:
                    chunk = chunk[:chunk_decompressed_bytes]
                chunks.append(chunk)

                if chunk_decompressed_bytes != chunk_size:
                    break

            out = b"".join(chunks)
        else:
            out = bytearray(size)
            decompressed_bytes = self.readinto(out)
            if decompressed_bytes != size:
                out = out[:decompressed_bytes]

        return out

    def close(self):
        if self._close_f_on_close:
            self.f.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class TextDecompressor(Decompressor):
    def read(self, *args, **kwargs) -> str:
        return super().read(*args, **kwargs).decode()


def decompress(data, *args, **kwargs):
    with BytesIO(data) as f:
        d = Decompressor(f, *args, **kwargs)
        return d.read()



from libcpp cimport bool
from libc.stdint cimport uint8_t, uint32_t

cdef extern from "tamp/common.h":
    ctypedef struct TampConf:
        int window
        int literal
        bool use_custom_dictionary
        # The lazy_matching field is conditionally compiled based on TAMP_LAZY_MATCHING
        # We declare it here, but accessing it when the macro is disabled will cause compile errors
        # This is handled in the Cython code by always setting it when the struct is initialized
        bool lazy_matching

    ctypedef enum tamp_res:
        # Normal/Recoverable status >= 0
        TAMP_OK = 0,
        TAMP_OUTPUT_FULL = 1,  # Wasn't able to complete action due to full output buffer.
        TAMP_INPUT_EXHAUSTED = 2, # Wasn't able to complete action due to exhausted input buffer.

        # Error Codes < 0
        TAMP_ERROR = -1,  # Generic error
        TAMP_EXCESS_BITS = -2,  # Provided symbol has more bits than conf->literal
        TAMP_INVALID_CONF = -3,  # Invalid configuration parameters.
        TAMP_OOB = -4,  # Out-of-bounds access detected in compressed data.

    void initialize_dictionary(unsigned char *buffer, size_t size, uint32_t seed);
    int compute_min_pattern_size(uint8_t window, uint8_t literal);


cdef extern from "tamp/compressor.h":
    ctypedef struct TampCompressor:
        pass

    tamp_res tamp_compressor_init(TampCompressor *compressor, const TampConf *conf, unsigned char *window);

    void tamp_compressor_sink(
            TampCompressor *compressor,
            const unsigned char *input,
            size_t input_size,
            size_t *consumed_size
            );

    tamp_res tamp_compressor_poll(
            TampCompressor *compressor,
            unsigned char *output,
            size_t output_size,
            size_t *output_written_size
            );

    tamp_res tamp_compressor_flush(
                    TampCompressor *compressor,
                    unsigned char *output,
                    size_t output_size,
                    size_t *output_written_size,
                    bool write_token
                    );

    tamp_res tamp_compressor_compress(
            TampCompressor *compressor,
            unsigned char *output,
            size_t output_size,
            size_t *output_written_size,
            const unsigned char *input,
            size_t input_size,
            size_t *input_consumed_size
            );


cdef extern from "tamp/decompressor.h":
    ctypedef struct TampDecompressor:
        pass

    tamp_res tamp_decompressor_read_header(
            TampConf *conf,
            const unsigned char *input,
            size_t input_size,
            size_t *input_consumed_size
    );

    tamp_res tamp_decompressor_init(
            TampDecompressor *decompressor,
            const TampConf *conf,
            unsigned char *window,
            uint8_t window_bits
    );

    tamp_res tamp_decompressor_decompress(
            TampDecompressor *decompressor,
            unsigned char *output,
            size_t output_size,
            size_t *output_written_size,
            const unsigned char *input,
            size_t input_size,
            size_t *input_consumed_size
    );



"""Micropython optimized for performance over readability.
"""
from io import BytesIO

import micropython
from micropython import const

from . import ExcessBitsError, bit_size, compute_min_pattern_size, initialize_dictionary

# encodes [2, 15] pattern lengths
_HUFFMAN_CODES = b"\x00\x03\x08\x0b\x14$&+KT\x94\x95\xaa'"
# These bit lengths pre-add the 1 bit for the 0-value is_literal flag.
_HUFFMAN_BITS = b"\x02\x03\x05\x05\x06\x07\x07\x07\x08\x08\x09\x09\x09\x07"
_FLUSH_CODE = const(0xAB)  # 8 bits


def _f_write(f, number, size):
    f.write(number.to_bytes(size, "big"))


class Compressor:
    def __init__(
        self,
        f,
        *,
        window=10,
        literal=8,
        dictionary=None,
    ):
        self._close_f_on_close = False
        if not hasattr(f, "write"):  # It's probably a path-like object.
            f = open(str(f), "wb")
            self._close_f_on_close = True

        self.window_bits = window
        self.literal_bits = literal

        self.min_pattern_size = compute_min_pattern_size(window, literal)

        # Window Buffer
        if dictionary:
            if bit_size(len(dictionary) - 1) != window:
                raise ValueError
            self.window_buf = dictionary
        else:
            self.window_buf = initialize_dictionary(1 << window)
        self.window_pos = 0

        # Input Buffer
        self.input_buf = bytearray(16)
        self.input_size = 0
        self.input_pos = 0

        # Write header
        self.f = f
        self.f_buf = ((window - 8) << 5 | (literal - 5) << 3 | int(bool(dictionary)) << 2) << (22)
        self.f_pos = 8

    @micropython.viper
    def _compress_input_buffer_single(self) -> int:
        bytes_written = 0
        # Viper-ize everything
        input_buf = ptr8(self.input_buf)
        input_size = int(self.input_size)
        input_pos = int(self.input_pos)

        literal_bits = int(self.literal_bits)
        literal_flag = 1 << literal_bits

        f_buf = int(self.f_buf)
        f_pos = int(self.f_pos)
        huffman_bits_ptr8 = ptr8(_HUFFMAN_BITS)
        huffman_codes_ptr8 = ptr8(_HUFFMAN_CODES)

        window_bits = int(self.window_bits)
        window_buf = ptr8(self.window_buf)
        window_size = 1 << int(window_bits)
        window_pos = int(self.window_pos)

        min_pattern_size = int(self.min_pattern_size)

        f = self.f

        # Find largest match
        search_i = int(0)
        match_size = int(0)

        if input_size >= min_pattern_size:
            for window_index in range(window_size - min_pattern_size + 1):
                if input_buf[input_pos] != window_buf[window_index]:
                    continue  # Significant speed short-cut

                input_index = (input_pos + 1) & 0xF
                if input_buf[input_index] != window_buf[window_index + 1]:
                    continue  # Small Speed short-cut

                current_match_size = int(2)
                for k in range(current_match_size, input_size):
                    input_index = (input_pos + k) & 0xF
                    if input_buf[input_index] != window_buf[window_index + k] or window_index + k >= window_size:
                        break
                    current_match_size = k + 1
                if current_match_size > match_size:
                    match_size = current_match_size
                    search_i = window_index

                    if match_size == input_size:
                        break

        # Write out a literal or a token
        if match_size >= min_pattern_size:
            huffman_index = match_size - min_pattern_size
            # Adds up to 9 bits
            f_pos += huffman_bits_ptr8[huffman_index]
            f_buf |= huffman_codes_ptr8[huffman_index] << (30 - f_pos)

            if f_pos >= 16:
                _f_write(f, f_buf >> 14, 2)
                f_buf = (f_buf & 0x3FFF) << 16
                f_pos -= 16
                bytes_written += 2

            # Adds up to 15 bits
            f_pos += window_bits
            f_buf |= search_i << (30 - f_pos)
        else:
            # Adds up to 9 bits
            match_size = 1
            if input_buf[input_pos] >> literal_bits:
                raise ExcessBitsError

            f_pos += literal_bits + 1
            f_buf |= (input_buf[input_pos] | literal_flag) << (30 - f_pos)

        if f_pos >= 16:
            _f_write(f, f_buf >> 14, 2)
            f_buf = (f_buf & 0x3FFF) << 16
            f_pos -= 16
            bytes_written += 2

        for _ in range(match_size):  # Copy pattern into buffer
            window_buf[window_pos] = input_buf[input_pos]
            input_pos = (input_pos + 1) & 0xF
            window_pos = (window_pos + 1) % window_size

        input_size -= match_size

        self.input_size = input_size
        self.input_pos = input_pos
        self.window_pos = window_pos

        self.f_pos = f_pos
        self.f_buf = f_buf

        return bytes_written

    @micropython.viper
    def write(self, data) -> int:
        bytes_written = 0

        data_l = int(len(data))
        data_p = ptr8(data)

        input_buf = ptr8(self.input_buf)

        max_pattern_size = int(self.min_pattern_size) + 13

        for i in range(data_l):
            input_size = int(self.input_size)
            input_pos = int(self.input_pos)

            pos = (input_pos + input_size) & 0xF
            input_buf[pos] = data_p[i]
            input_size += 1
            self.input_size = input_size

            if input_size == max_pattern_size:
                bytes_written += int(self._compress_input_buffer_single())
        return bytes_written

    def flush(self, write_token=True):
        bytes_written = 0
        while self.input_size:
            bytes_written += self._compress_input_buffer_single()

        if self.f_pos > 0 and write_token:
            self.f_pos += 9
            self.f_buf |= _FLUSH_CODE << (30 - self.f_pos)

        while self.f_pos > 0:
            _f_write(self.f, self.f_buf >> 22, 1)
            self.f_buf = (self.f_buf & 0x3FFFFF) << 8
            self.f_pos -= 8
            bytes_written += 1

        self.f_pos = 0

        return bytes_written

    def close(self):
        bytes_written = self.flush(write_token=False)
        if self._close_f_on_close:
            self.f.close()
        return bytes_written

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class TextCompressor(Compressor):
    def write(self, data: str) -> int:
        return super().write(data.encode())


def compress(data, *args, **kwargs):
    with BytesIO() as f:
        cls = TextCompressor if isinstance(data, str) else Compressor
        c = cls(f, *args, **kwargs)
        c.write(data)
        c.flush(write_token=False)
        f.seek(0)
        return f.read()



import sys
from pathlib import Path
from typing import Annotated, Literal, Optional

from cyclopts import App, Parameter, validators

import tamp

app = App(help="Compress/Decompress data in Tamp format.")

ImplementationType = Optional[Literal["c", "python"]]


def get_compress_implementation(impl_name: Optional[str]):
    """Get the compress function based on the implementation name.

    Parameters
    ----------
    impl_name : Optional[str]
        Name of the implementation ('c', 'python', or None for default)

    Returns
    -------
    callable
        The requested compress function
    """
    if impl_name is None:
        # Use default auto-detection
        return tamp.compress

    impl_name = impl_name.lower()

    if impl_name == "c":
        try:
            from tamp._c_compressor import compress

            return compress
        except ImportError:
            raise ImportError(
                "C implementation not available. Please ensure the C extensions are compiled "
                "by running: poetry run python build.py build_ext --inplace"
            )
    elif impl_name == "python":
        from tamp.compressor import compress

        return compress
    else:
        raise ValueError(f"Unknown implementation: {impl_name}. Valid options are 'c' or 'python'")


def get_decompress_implementation(impl_name: Optional[str]):
    """Get the decompress function based on the implementation name.

    Parameters
    ----------
    impl_name : Optional[str]
        Name of the implementation ('c', 'python', or None for default)

    Returns
    -------
    callable
        The requested decompress function
    """
    if impl_name is None:
        # Use default auto-detection
        return tamp.decompress

    impl_name = impl_name.lower()

    if impl_name == "c":
        try:
            from tamp._c_decompressor import decompress

            return decompress
        except ImportError:
            raise ImportError(
                "C implementation not available. Please ensure the C extensions are compiled "
                "by running: poetry run python build.py build_ext --inplace"
            )
    elif impl_name == "python":
        from tamp.decompressor import decompress

        return decompress
    else:
        raise ValueError(f"Unknown implementation: {impl_name}. Valid options are 'c' or 'python'")


def read(input_: Optional[Path]) -> bytes:
    data = sys.stdin.buffer.read() if input_ is None else input_.read_bytes()
    if not data:
        raise ValueError("No data provided.")
    return data


def write(output: Optional[Path], data: bytes):
    if output is None:
        sys.stdout.buffer.write(data)
    else:
        output.write_bytes(data)


@app.command()
def compress(
    input_: Annotated[Optional[Path], Parameter(name=["--input", "-i"])] = None,
    output: Annotated[Optional[Path], Parameter(name=["--output", "-o"])] = None,
    *,
    window: Annotated[
        int,
        Parameter(
            name=["--window", "-w"],
            validator=validators.Number(gte=8, lte=15),
        ),
    ] = 10,
    literal: Annotated[
        int,
        Parameter(
            name=["--literal", "-l"],
            validator=validators.Number(gte=5, lte=8),
        ),
    ] = 8,
    lazy_matching: bool = False,
    implementation: ImplementationType = None,
):
    """Compress an input file or stream.

    Parameters
    ----------
    input_: Optional[Path]
        Input file to compress. Defaults to stdin.
    output: Optional[Path]
        Output compressed file. Defaults to stdout.
    window: int
        Number of bits used to represent the dictionary window.
    literal: int
        Number of bits used to represent a literal.
    lazy_matching: bool
        Use roughly 50% more cpu to get 0~2% better compression.
    implementation: Optional[Literal["c", "python"]]
        Explicitly specify which implementation to use (c or python). Defaults to auto-detection.
    """
    input_bytes = read(input_)
    compress_fn = get_compress_implementation(implementation)
    output_bytes = compress_fn(
        input_bytes,
        window=window,
        literal=literal,
        lazy_matching=lazy_matching,
    )
    write(output, output_bytes)


@app.command()
def decompress(
    input_: Annotated[Optional[Path], Parameter(name=["--input", "-i"])] = None,
    output: Annotated[Optional[Path], Parameter(name=["--output", "-o"])] = None,
    *,
    implementation: ImplementationType = None,
):
    """Decompress an input file or stream.

    Parameters
    ----------
    input_: Optional[Path]
        Input file to decompress. Defaults to stdin.
    output: Optional[Path]
        Output decompressed file. Defaults to stdout.
    implementation: Optional[Literal["c", "python"]]
        Explicitly specify which implementation to use (c or python). Defaults to auto-detection.
    """
    input_bytes = read(input_)
    decompress_fn = get_decompress_implementation(implementation)
    output_bytes = decompress_fn(input_bytes)
    write(output, output_bytes)


def run_app():
    app()



import builtins
cimport ctamp
from libcpp cimport bool
from cpython.exc cimport PyErr_CheckSignals
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.stddef cimport size_t
from io import BytesIO
from . import bit_size
from ._c_common cimport CHUNK_SIZE
from ._c_common import ERROR_LOOKUP

from typing import Union


cdef class Compressor:
    cdef ctamp.TampCompressor* _c_compressor
    cdef bytearray _window_buffer
    cdef unsigned char *_window_buffer_ptr
    cdef object f
    cdef bint _close_f_on_close

    def __cinit__(self):
        self._c_compressor = <ctamp.TampCompressor*>PyMem_Malloc(sizeof(ctamp.TampCompressor))
        if self._c_compressor is NULL:
            raise MemoryError

    def __dealloc__(self):
        PyMem_Free(self._c_compressor)

    def __init__(
        self,
        f,
        *,
        int window=10,
        int literal=8,
        dictionary=None,
        bool lazy_matching=False,
    ):
        cdef ctamp.TampConf conf

        if dictionary and bit_size(len(dictionary) - 1) != window:
            raise ValueError("Dictionary-window size mismatch.")

        if not hasattr(f, "write"):  # It's probably a path-like object.
            f = open(str(f), "wb")
            self._close_f_on_close = True
        else:
            self._close_f_on_close = False

        self.f = f

        conf.window = window
        conf.literal = literal
        conf.use_custom_dictionary = builtins.bool(dictionary)
        # Set lazy_matching - this field is conditionally compiled based on TAMP_LAZY_MATCHING
        # The build system defines this macro, so the field should be available
        conf.lazy_matching = lazy_matching

        self._window_buffer = dictionary if dictionary else bytearray(1 << window)
        self._window_buffer_ptr = <unsigned char *>self._window_buffer

        res = ctamp.tamp_compressor_init(self._c_compressor, &conf, self._window_buffer_ptr)
        if res < 0:
            raise ERROR_LOOKUP.get(res, NotImplementedError)

    def write(self, const unsigned char[::1] data not None) -> int:
        cdef:
            ctamp.tamp_res res

            bytearray output_buffer = bytearray(CHUNK_SIZE)
            unsigned char *output_buffer_ptr = output_buffer
            const unsigned char* data_ptr = &data[0]

            size_t input_consumed_size = 0
            size_t input_remaining_size = data.shape[0]
            size_t output_buffer_written_size
            int written_to_disk_size = 0

        output_buffer_mv = memoryview(output_buffer)

        while input_remaining_size:
            res = ctamp.tamp_compressor_compress(
                self._c_compressor,
                output_buffer_ptr,
                CHUNK_SIZE,
                &output_buffer_written_size,
                data_ptr,
                input_remaining_size,
                &input_consumed_size
            )
            if res < 0:
                raise ERROR_LOOKUP.get(res, NotImplementedError)
            self.f.write(output_buffer_mv[:output_buffer_written_size])
            written_to_disk_size += output_buffer_written_size
            data_ptr += input_consumed_size
            input_remaining_size -= input_consumed_size

            # Check signals for things like KeyboardInterrupt
            PyErr_CheckSignals()

        return written_to_disk_size

    cpdef int flush(self, bool write_token = True) except -1:
        cdef ctamp.tamp_res res
        cdef bytearray buffer = bytearray(24)
        cdef size_t output_written_size = 0

        res = ctamp.tamp_compressor_flush(
            self._c_compressor,
            <unsigned char *> buffer,
            len(buffer),
            &output_written_size,
            write_token,
        )

        if res < 0:
            raise ERROR_LOOKUP.get(res, NotImplementedError)

        if output_written_size:
            self.f.write(buffer[:output_written_size])

        self.f.flush()

        return output_written_size

    def close(self) -> int:
        bytes_written = self.flush(write_token=False)
        if self._close_f_on_close:
            self.f.close()
        return bytes_written

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class TextCompressor(Compressor):

    def write(self, data: str) -> int:
        return super().write(data.encode())


def compress(data: Union[bytes, str], *args, **kwargs) -> bytes:
    with BytesIO() as f:
        if isinstance(data, str):
            c = TextCompressor(f, *args, **kwargs)
            c.write(data)
        else:
            c = Compressor(f, *args, **kwargs)
            c.write(data)
        c.flush(write_token=False)
        f.seek(0)
        return f.read()



__version__ = "1.11.1"


class ExcessBitsError(Exception):
    """Provided data has more bits than expected ``literal`` bits."""


def bit_size(value):
    for i in range(32):
        if not value:
            return i
        value >>= 1
    return -1


def _xorshift32(seed):
    while True:
        seed ^= (seed << 13) & 0xFFFFFFFF
        seed ^= (seed >> 17) & 0xFFFFFFFF
        seed ^= (seed << 5) & 0xFFFFFFFF
        yield seed


def initialize_dictionary(source, seed=None):
    if seed is None:
        seed = 3758097560
    elif seed == 0:
        return bytearray(source)

    out = source if isinstance(source, bytearray) else bytearray(source)
    size = len(out)

    chars = b" \x000ei>to<ans\nr/."  # 16 most common chars in dataset

    def _gen_stream(xorshift32):
        for _ in range(size >> 3):
            value = next(xorshift32)
            for _ in range(8):
                yield chars[value & 0x0F]
                value >>= 4

    for i, c in enumerate(_gen_stream(_xorshift32(seed))):
        out[i] = c

    return out


def compute_min_pattern_size(window, literal):
    if not (7 < window < 16 and 4 < literal < 9):
        raise ValueError

    return 2 + (window > (10 + ((literal - 5) << 1)))


try:
    from .compressor_viper import Compressor, TextCompressor, compress
except ImportError:
    try:
        from ._c_compressor import Compressor, TextCompressor, compress
    except ImportError:
        try:
            from .compressor import Compressor, TextCompressor, compress
        except ImportError:
            pass

try:
    from .decompressor_viper import Decompressor, TextDecompressor, decompress
except ImportError:
    try:
        from ._c_decompressor import Decompressor, TextDecompressor, decompress
    except ImportError:
        try:
            from .decompressor import Decompressor, TextDecompressor, decompress
        except ImportError:
            pass


def open(f, mode="rb", **kwargs):
    if "r" in mode and "w" in mode:
        raise ValueError

    if "r" in mode:  # Decompressor
        return Decompressor(f, **kwargs) if "b" in mode else TextDecompressor(f, **kwargs)
    elif "w" in mode:  # Compressor
        return Compressor(f, **kwargs) if "b" in mode else TextCompressor(f, **kwargs)
    else:
        raise ValueError



cimport ctamp
from . import ExcessBitsError

cdef int CHUNK_SIZE = (1 << 20)

ERROR_LOOKUP = {
    # Recoverable Errors
    ctamp.TAMP_OUTPUT_FULL: IndexError,
    ctamp.TAMP_INPUT_EXHAUSTED: IndexError,

    # Bad Errors
    ctamp.TAMP_ERROR: Exception,
    ctamp.TAMP_EXCESS_BITS: ExcessBitsError,
    ctamp.TAMP_INVALID_CONF: ValueError,
    ctamp.TAMP_OOB: ValueError,
}



cimport ctamp
from cpython.exc cimport PyErr_CheckSignals
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.stddef cimport size_t
from io import BytesIO
from . import bit_size
from ._c_common cimport CHUNK_SIZE
from ._c_common import ERROR_LOOKUP


from typing import Union


cdef class Decompressor:
    cdef:
        ctamp.TampDecompressor* _c_decompressor
        bytearray _window_buffer
        unsigned char *_window_buffer_ptr
        object input_buffer
        unsigned char *input_buffer_ptr
        size_t input_size
        size_t input_consumed
        object f
        cdef bint _close_f_on_close

    def __cinit__(self):
        self._c_decompressor = <ctamp.TampDecompressor*>PyMem_Malloc(sizeof(ctamp.TampDecompressor))
        if self._c_decompressor is NULL:
            raise MemoryError

    def __dealloc__(self):
        PyMem_Free(self._c_decompressor)

    def __init__(self, f, *, dictionary=None):
        cdef:
            ctamp.TampConf conf
            size_t input_consumed_size;

        if not hasattr(f, "read"):  # It's probably a path-like object.
            f = open(str(f), "rb")
            self._close_f_on_close = True
        else:
            self._close_f_on_close = False

        self.f = f

        self.input_buffer = bytearray(CHUNK_SIZE)
        self.input_buffer_ptr = self.input_buffer
        self.input_size = 0
        self.input_consumed = 0

        compressed_data = f.read(1)

        res = ctamp.tamp_decompressor_read_header(&conf, compressed_data, len(compressed_data), &input_consumed_size);
        if res != ctamp.TAMP_OK:
            raise ERROR_LOOKUP.get(res, NotImplementedError)
        if conf.use_custom_dictionary and dictionary is None:
            raise ValueError

        self._window_buffer = dictionary if dictionary else bytearray(1 << conf.window)
        self._window_buffer_ptr = <unsigned char *>self._window_buffer

        res = ctamp.tamp_decompressor_init(self._c_decompressor, &conf, self._window_buffer_ptr, conf.window)
        if res < 0:
            raise ERROR_LOOKUP.get(res, NotImplementedError)

    def readinto(self, bytearray buf) -> int:
        cdef:
            unsigned char *output_buffer_ptr = buf
            size_t output_written_size, input_chunk_consumed

        size = len(buf)
        while size:
            # decompress in chunks so we can skill check PyErr_CheckSignals
            output_size = min(CHUNK_SIZE, size)

            res = ctamp.tamp_decompressor_decompress(
                self._c_decompressor,
                output_buffer_ptr,
                output_size,
                &output_written_size,
                self.input_buffer_ptr + self.input_consumed,
                self.input_size,
                &input_chunk_consumed
            )
            self.input_size -= input_chunk_consumed
            self.input_consumed += input_chunk_consumed
            output_buffer_ptr += output_written_size
            size -= output_written_size

            if res == ctamp.TAMP_INPUT_EXHAUSTED:
                # Read in more data
                self.input_size = self.f.readinto(self.input_buffer)
                self.input_consumed = 0
                if self.input_size == 0:
                    break;
            elif res < 0:
                raise ERROR_LOOKUP.get(res, NotImplementedError)

            # Check signals for things like KeyboardInterrupt
            PyErr_CheckSignals()

        return len(buf) - size

    def read(self, int size = -1) -> bytearray:
        """
        Parameters
        ----------
        size: int
            Number of bytes to attempt to read in.
            If negative 1, read until end of stream.
        """
        if size == 0:
            return bytearray()

        chunk_size = CHUNK_SIZE
        out = []
        while True:
            # buf holds decompressed data.
            buf = bytearray(chunk_size if size < 0 else size)
            chunk_size <<= 1  # Keep allocating larger chunks as we go on.
            # decompression happens inside of readinto
            read_size = self.readinto(buf)
            if size > 0:
                # We one-shot the decompression
                out.append(buf)
                break
            # We are decompressing in chunks
            if read_size < len(buf):
                if read_size:
                    out.append(buf[:read_size])
                break
            else:
                out.append(buf)
        return out[0] if len(out) == 1 else bytearray(b"".join(out))

    def close(self):
        if self._close_f_on_close:
            self.f.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

class TextDecompressor(Decompressor):
    def read(self, *args, **kwargs) -> str:
        return super().read(*args, **kwargs).decode()


def decompress(data: bytes, *args, **kwargs) -> bytearray:
    with BytesIO(data) as f:
        d = Decompressor(f, *args, **kwargs)
        return d.read()



from io import BytesIO

try:
    from typing import Optional
except ImportError:
    pass

from . import compute_min_pattern_size, initialize_dictionary

_CHUNK_SIZE = 1 << 20
_FLUSH = object()

# Each key here are the huffman codes or'd with 0x80
# This is so that each lookup is easy/quick.
_huffman_lookup = {
    0b0: 0,
    0b11: 1,
    0b1000: 2,
    0b1011: 3,
    0b10100: 4,
    0b100100: 5,
    0b100110: 6,
    0b101011: 7,
    0b1001011: 8,
    0b1010100: 9,
    0b10010100: 10,
    0b10010101: 11,
    0b10101010: 12,
    0b100111: 13,
    0b10101011: _FLUSH,
}


class _BitReader:
    """Reads bits from a stream."""

    def __init__(self, f, close_f_on_close=False):
        self.close_f_on_close = close_f_on_close
        self.f = f
        self.clear()

    def read_huffman(self):
        proposed_code = 0
        lookup = _huffman_lookup
        read = self.read
        for _ in range(8):
            proposed_code |= read(1)
            try:
                return lookup[proposed_code]
            except KeyError:
                proposed_code <<= 1
        raise RuntimeError("Unable to decode huffman code. Should never happen.")

    def read(self, num_bits):
        while self.bit_pos < num_bits:
            byte = self.f.read(1)
            if not byte:
                raise EOFError
            byte_value = int.from_bytes(byte, "little")
            self.buffer |= byte_value << (24 - self.bit_pos)
            self.bit_pos += 8

            if self.backup_buffer is not None and self.backup_bit_pos is not None:
                self.backup_buffer |= byte_value << (24 - self.backup_bit_pos)
                self.backup_bit_pos += 8

        result = self.buffer >> (32 - num_bits)
        mask = (1 << (32 - num_bits)) - 1
        self.buffer = (self.buffer & mask) << num_bits
        self.bit_pos -= num_bits

        return result

    def clear(self):
        self.buffer = 0
        self.bit_pos = 0

        self.backup_buffer = None
        self.backup_bit_pos = None

    def close(self):
        if self.close_f_on_close:
            self.f.close()

    def __len__(self):
        return self.bit_pos

    def __enter__(self):
        # Context manager to restore all read bits in a session if any read fails.

        # backup the buffer
        self.backup_buffer = self.buffer
        self.backup_bit_pos = self.bit_pos
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            # restore buffers
            self.buffer = self.backup_buffer
            self.bit_pos = self.backup_bit_pos

        self.backup_buffer = None
        self.backup_bit_pos = None


class _RingBuffer:
    def __init__(self, buffer):
        self.buffer = buffer
        self.size = len(buffer)
        self.pos = 0  # Always pointing to the byte-to-be-overwritten
        self.index = self.buffer.index

    def write_byte(self, byte):  # ~10% of time
        self.buffer[self.pos] = byte
        self.pos += 1
        if self.pos == self.size:
            self.pos = 0

    def write_bytes(self, data):
        for byte in data:
            self.write_byte(byte)


class Decompressor:
    """Decompresses a file or stream of tamp-compressed data.

    Can be used as a context manager to automatically handle file
    opening and closing:

    .. code-block:: python

        with tamp.Decompressor("compressed.tamp") as f:
            decompressed_data = f.read()
    """

    def __init__(self, f, *, dictionary: Optional[bytearray] = None):
        """
        Parameters
        ----------
        f: Union[file, str]
            File-like object to read compressed bytes from.
        dictionary: Optional[bytearray]
            Use the given **initialized** buffer inplace.
            At compression time, the same initialized buffer must be provided.
            Decompression stream's ``window`` must agree with the dictionary size.
            If providing a pre-allocated buffer, but with default initialization, it must
            first be initialized with :func:`~tamp.initialize_dictionary`
        """
        if not hasattr(f, "read"):  # It's probably a path-like object.
            f = open(str(f), "rb")
            close_f_on_close = True
        else:
            close_f_on_close = False

        self._bit_reader = _BitReader(f, close_f_on_close=close_f_on_close)

        # Read Header
        self.window_bits = self._bit_reader.read(3) + 8
        self.literal_bits = self._bit_reader.read(2) + 5
        uses_custom_dictionary = self._bit_reader.read(1)
        reserved = self._bit_reader.read(1)
        more_header_bytes = self._bit_reader.read(1)

        if reserved:
            raise NotImplementedError

        if more_header_bytes:
            raise NotImplementedError

        if uses_custom_dictionary and dictionary is None:
            raise ValueError

        self._window_buffer = _RingBuffer(
            buffer=(dictionary if dictionary else initialize_dictionary(1 << self.window_bits)),
        )

        self.min_pattern_size = compute_min_pattern_size(self.window_bits, self.literal_bits)

        self.overflow = bytearray()

    def readinto(self, buf: bytearray) -> int:
        """Decompresses data into provided buffer.

        Parameters
        ----------
        buf: bytearray
            Buffer to decode data into.

        Returns
        -------
        int
            Number of bytes decompressed into buffer.
        """
        if len(self.overflow) > len(buf):
            buf[:] = self.overflow[: len(buf)]
            written = len(buf)
            self.overflow = self.overflow[len(buf) :]
            return written
        elif self.overflow:
            buf[: len(self.overflow)] = self.overflow
            written = len(self.overflow)
            self.overflow = bytearray()
        else:
            written = 0

        while written < len(buf):
            try:
                with self._bit_reader:
                    is_literal = self._bit_reader.read(1)

                    if is_literal:
                        c = self._bit_reader.read(self.literal_bits)
                        self._window_buffer.write_byte(c)
                        buf[written] = c
                        written += 1
                    else:
                        match_size = self._bit_reader.read_huffman()
                        if match_size is _FLUSH:
                            self._bit_reader.clear()
                            continue
                        match_size += self.min_pattern_size
                        index = self._bit_reader.read(self.window_bits)

                        string = self._window_buffer.buffer[index : index + match_size]
                        self._window_buffer.write_bytes(string)

                        to_buf = min(len(buf) - written, match_size)
                        buf[written : written + to_buf] = string[:to_buf]
                        written += to_buf
                        if to_buf < match_size:
                            self.overflow[:] = string[to_buf:]
                            break
            except EOFError:
                break

        return written

    def read(self, size: int = -1) -> bytearray:
        """Decompresses data to bytes.

        Parameters
        ----------
        size: int
            Maximum number of bytes to return.
            If a negative value is provided, all data will be returned.
            Defaults to ``-1``.

        Returns
        -------
        bytearray
            Decompressed data.
        """
        if size == 0:
            return bytearray()

        chunk_size = _CHUNK_SIZE
        out = []
        while True:
            buf = bytearray(chunk_size if size < 0 else size)
            chunk_size <<= 1  # Keep allocating larger chunks as we go on.
            read_size = self.readinto(buf)
            if size > 0:
                # Read the entire contents in one go.
                out.append(buf)
                break
            else:
                if read_size < len(buf):
                    if read_size:
                        out.append(buf[:read_size])
                    break
                else:
                    out.append(buf)
        return out[0] if len(out) == 1 else bytearray(b"".join(out))

    def close(self):
        """Closes the input file or stream, if tamp opened it."""
        self._bit_reader.close()

    def __enter__(self):
        """Use :class:`Decompressor` as a context manager.

        .. code-block:: python

           with tamp.Decompressor("output.tamp") as f:
               decompressed_data = f.read()
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Calls :meth:`~Decompressor.close` on contextmanager exit."""
        self.close()


class TextDecompressor(Decompressor):
    """Decompresses a file or stream of tamp-compressed data into text."""

    def read(self, size: int = -1) -> str:
        """Decompresses data to text.

        Parameters
        ----------
        size: int
            Maximum number of bytes to return.
            If a negative value is provided, all data will be returned.
            Defaults to ``-1``.

        Returns
        -------
        str
            Decompressed text.
        """
        return super().read(size).decode()


def decompress(data: bytes, *, dictionary: Optional[bytearray] = None) -> bytearray:
    """Single-call to decompress data.

    Parameters
    ----------
    data: bytes
        Tamp-compressed data to decompress.
    dictionary: Optional[bytearray]
        Use the given **initialized** buffer inplace.
        At compression time, the same initialized buffer must be provided.
        Decompression stream's ``window`` must agree with the dictionary size.
        If providing a pre-allocated buffer, but with default initialization, it must
        first be initialized with :func:`~tamp.initialize_dictionary`

    Returns
    -------
    bytearray
        Decompressed data.
    """
    with BytesIO(data) as f:
        d = Decompressor(f, dictionary=dictionary)
        return d.read()



from __future__ import annotations

from pathlib import Path
from typing import Literal, Optional, Union, overload

from .compressor import Compressor as Compressor
from .compressor import TextCompressor as TextCompressor
from .compressor import compress as compress
from .decompressor import Decompressor as Decompressor
from .decompressor import TextDecompressor as TextDecompressor
from .decompressor import decompress as decompress

__version__: str

PathLike = Path | str

OpenTextModeWriting = Literal["w", "wt", "tw"]
OpenTextModeReading = Literal["r", "rt", "tr"]
OpenTextMode = OpenTextModeWriting | OpenTextModeReading

OpenBinaryModeWriting = Literal["wb", "bw"]
OpenBinaryModeReading = Literal["rb", "br"]
OpenBinaryMode = OpenBinaryModeReading | OpenBinaryModeWriting

OpenModeWriting = OpenTextModeWriting | OpenBinaryModeWriting
OpenModeReading = OpenTextModeReading | OpenBinaryModeReading

@overload
def open(
    f: PathLike,
    mode: OpenBinaryModeReading,
    **kwargs,
) -> Decompressor: ...
@overload
def open(
    f: PathLike,
    mode: OpenTextModeReading,
    **kwargs,
) -> TextDecompressor: ...
@overload
def open(
    f: PathLike,
    mode: OpenBinaryModeWriting,
    **kwargs,
) -> Compressor: ...
@overload
def open(
    f: PathLike,
    mode: OpenTextModeWriting,
    **kwargs,
) -> TextCompressor: ...
def bit_size(value: int) -> int: ...
def initialize_dictionary(size: int, seed: int | None = None) -> bytearray: ...
def compute_min_pattern_size(window: int, literal: int) -> int: ...

class ExcessBitsError(Exception):
    """Provided data has more bits than expected ``literal`` bits."""
